-- UART send
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity uart_send is
	generic(
		TCQ            : time := 2 ns;
		DATA_WIDTH     : natural := 32;
		BIT_CLOCK_TIME : natural := 4
	);
	port(
		clk_in     : in  std_logic;
		reset_in   : in  std_logic;
		cs_in      : in  std_logic;
		a0_in      : in  std_logic;
		wr_in      : in  std_logic;
		rd_in      : in  std_logic;
		data_in    : in  std_logic_vector(DATA_WIDTH - 1 downto 0);
		data_out   : out std_logic_vector(DATA_WIDTH - 1 downto 0);
		serial_out : out std_logic
	);

end entity;

architecture rtl of uart_send is
	constant BYTE_WIDTH   : natural := 8;
	constant NUM_OF_WORDS : natural := DATA_WIDTH / BYTE_WIDTH;
	type state_type is (ready, wait_data, take_byte, send_byte, end_transfer);
	type data_reg_type is array (0 to NUM_OF_WORDS - 1) of std_logic_vector(BYTE_WIDTH - 1 downto 0);
	subtype reg_type is std_logic_vector(DATA_WIDTH - 1 downto 0);
	signal state_reg, state_next       : state_type;
	signal data_reg, data_next         : reg_type;
	signal ctrl_reg, ctrl_next         : reg_type;
	signal status_reg, status_next     : reg_type;
	signal byte_cnt_reg, byte_cnt_next : natural range 0 to NUM_OF_WORDS;
	signal bit_cnt_reg, bit_cnt_next   : natural range 0 to BYTE_WIDTH + 1;
	signal bit_time_reg, bit_time_next : natural range 0 to BIT_CLOCK_TIME;
	signal send_reg, send_next         : unsigned(0 to BYTE_WIDTH + 1);

begin
	process(clk_in, reset_in) is
	begin
		if reset_in = '1' then
			state_reg    <= ready;
			data_reg     <= (others => '0');
			ctrl_reg     <= (others => '0');
			status_reg   <= (others => '0');
			byte_cnt_reg <= 0;
			bit_cnt_reg  <= 0;
			bit_time_reg <= 0;
			send_reg     <= (others => '0');
		elsif (clk_in'event and clk_in = '1') then
			state_reg    <= state_next after TCQ;
			data_reg     <= data_next after TCQ;
			ctrl_reg     <= ctrl_next after TCQ;
			status_reg   <= status_next after TCQ;
			byte_cnt_reg <= byte_cnt_next after TCQ;
			bit_cnt_reg  <= bit_cnt_next after TCQ;
			bit_time_reg <= bit_time_next after TCQ;
			send_reg     <= send_next after TCQ;
		end if;
	end process;

	--Next-state calculation
	process(state_reg, cs_in, a0_in, wr_in, data_reg, ctrl_reg, status_reg, byte_cnt_reg, bit_cnt_reg, bit_time_reg, send_reg, data_in) is
		variable byte_data : data_reg_type;
	begin
		state_next    <= state_reg;
		data_next     <= data_reg;
		ctrl_next     <= ctrl_reg;
		status_next   <= status_reg;
		byte_cnt_next <= byte_cnt_reg;
		bit_cnt_next  <= bit_cnt_reg;
		bit_time_next <= bit_time_reg;
		send_next     <= send_reg;
		serial_out    <= '0';
		case state_reg is
			when ready =>
				if (ctrl_reg(0) = '1') then
					status_next(1 downto 0) <= "01"; --ready to wait data and no error
					state_next     <= wait_data;
				elsif (cs_in = '1' and wr_in = '1' and a0_in = '1') then
					ctrl_next   <= data_in;
					status_next <= (others => '0');
				end if;
			when wait_data =>
				if (cs_in = '1' and wr_in = '1' and a0_in = '0') then
					data_next      <= data_in;
					byte_cnt_next  <= 0;
					status_next(0) <= '0';
					state_next     <= take_byte;
				end if;
			when take_byte =>
				for i in byte_data'range loop
					for j in byte_data(i)'range loop
						byte_data(i)(j) := data_reg(i * BYTE_WIDTH + j);
					end loop;
				end loop;
				send_next(1 to BYTE_WIDTH) <= unsigned(byte_data(byte_cnt_reg));
				send_next(0)               <= '1';
				send_next(BYTE_WIDTH + 1)  <= '1';
				byte_cnt_next              <= byte_cnt_reg + 1;
				bit_cnt_next               <= BYTE_WIDTH + 1;
				bit_time_next              <= BIT_CLOCK_TIME - 1;
				state_next                 <= send_byte;
			when send_byte =>
				serial_out <= send_reg(BYTE_WIDTH + 1);
				if (bit_time_reg = 0) then
					if (bit_cnt_reg = 0) then
						if (byte_cnt_reg = NUM_OF_WORDS) then --finish transfer
							state_next <= end_transfer;
						else            --next byte
							state_next <= take_byte;
						end if;
					else                -- send next bit
						bit_cnt_next  <= bit_cnt_reg - 1;
						bit_time_next <= BIT_CLOCK_TIME - 1;
						send_next     <= shift_right(send_reg, 1);
					end if;
				else                    -- send current bit
					bit_time_next <= bit_time_reg - 1;
				end if;
			when end_transfer =>
				status_next(0) <= '1';
				ctrl_next(0)   <= '0';
				state_next     <= ready;
		end case;
	end process;

	-- Calculating data output
	process(cs_in, a0_in, rd_in, data_reg, status_reg)
	begin
		if (rd_in = '1') and (cs_in = '1') then
			if (a0_in = '0') then
				data_out <= data_reg;
			else
				data_out <= status_reg;
			end if;
		else
			data_out <= (others => '0');
		end if;

	end process;

end rtl;
